------------------------------------------------------------------------------
------------------------------------------------------------------------------
--                                                                          --
-- Copyright (c) 2008-2009 Tobias Gubener <tobiflex@opencores.org>          --
--                                                                          --
-- This source file is free software: you can redistribute it and/or modify --
-- it under the terms of the GNU General Public License as published        --
-- by the Free Software Foundation, either version 3 of the License, or     --
-- (at your option) any later version.                                      --
--                                                                          --
-- This source file is distributed in the hope that it will be useful,      --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of           --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            --
-- GNU General Public License for more details.                             --
--                                                                          --
-- You should have received a copy of the GNU General Public License        --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.    --
--                                                                          --
------------------------------------------------------------------------------
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;


entity acex is
   port ( 
	addonin: in std_logic_vector(8 downto 0);	
	C2MHz: in std_logic;	
	ioshift: in std_logic;	
	kb_clk: inout std_logic;	
	kb_data: inout std_logic;	
	ms_clk: inout std_logic;	
	ms_data: inout std_logic;	
	addonout: out std_logic_vector(5 downto 0);		
	
	idedata: inout std_logic_vector(15 downto 0);		
	idea: out std_logic_vector(2 downto 0);		
	ide_wr: out std_logic;			
	ide_rd: buffer std_logic;			
	ide_csp0: buffer std_logic;			
	ide_css0: buffer std_logic;
	TxD: out std_logic;
	lrclk: out std_logic;
	adat: out std_logic;
		
	ring_o : out std_logic_vector(3 downto 0)
			
	
  );

end acex;


architecture wire of acex is

	signal shift: std_logic_vector(9 downto 0);
	signal Ring : std_logic_vector(3 downto 0);
	signal ideoutd: std_logic_vector(15 downto 0);
	signal ideoutdd: std_logic_vector(12 downto 0);
	signal ideind: std_logic_vector(10 downto 0);
	signal kb_clki : std_logic;
	signal kb_datai : std_logic;
	signal ms_clki : std_logic;
	signal ms_datai : std_logic;
	signal ide_wrin : std_logic;
	signal ide_wrd : std_logic;
	signal C2MHzd : std_logic;
	signal sync: std_logic;	
	signal inbuf : std_logic_vector(4 downto 0);

--type micro_states is (idle, nop, io_aktion, io_rd_data, ide1, ide2, ide3, ide4, ide5, ide6, ide7, addon1, addon2, addon3, addon4, addon5,
--						conf1, conf2, conf3, conf4, conf5, conf6, conf7, conf8, conf9, conf10, conf11, conf12, conf13, conf14, conf15, conf16);
--signal micro_state		: micro_states;
--signal next_micro_state		: micro_states;

-------------------------------------------------------------------			
-------------------------------------------------------------------		
BEGIN
	ring_o <= Ring;
	idedata <= ideoutd WHEN ide_wrd='0' ELSE "ZZZZZZZZZZZZZZZZ";


	kb_clk <= '0' WHEN kb_clki='0' ELSE 'Z'; 
	kb_data <= '0' WHEN kb_datai='0' ELSE 'Z'; 
	ms_clk <= '0' WHEN ms_clki='0' ELSE 'Z'; 
	ms_data <= '0' WHEN ms_datai='0' ELSE 'Z'; 
--	ide_wr <= '0' WHEN ide_wrin='0' AND ide_wrd='0' ELSE '1';
	ide_wr <= '0' WHEN ide_wrin='0' ELSE '1';
-------------------------------------------------------------------	
--  RESEIVER		
--			0			1			2			3     

--addonin	1			2			3			0
--(0)								TxD		 	sync
--(1)		ide_rd 					ide_csp0	kb/ms_datai
--(2)		ide_wr					ide_css0 	kb/ms_clki	
--(3)		   sysclk
--(4)		idea(0)		idei(0),shift	idei(5)		idei(10)	
--(5)		idea(1) 	idei(1),lrclk	idei(6)		idei(11)
--(6)		idea(2) 	idei(2)		idei(7)		idei(12)
--(7)		idea(3) 	idei(3)		idei(8)		idei(13)
--(8)		idei(15)	idei(4)		idei(9)		idei(14)
	
	PROCESS(addonin)
	BEGIN
		IF addonin(3)'event AND addonin(3) = '1' THEN
			inbuf <= inbuf(3 downto 0)&addonin(0);
			Ring(0) <= NOT (Ring(0) OR Ring(1) OR Ring(2));	--only one hot bit
			IF Ring(1)='0' OR inbuf(4)/=inbuf(0) THEN
				Ring(1) <= Ring(0);
				Ring(2) <= Ring(1);
			END IF;
			Ring(3) <= Ring(2);
		END IF;
		
		IF addonin(3)'event AND addonin(3) = '1' THEN
			IF Ring(0)='1' THEN
--				ide_rd <= addonin(2);
				IF ide_wrd='1' AND ide_wrin='0' THEN 
					ideoutd(9 downto 0) <= ideoutdd(9 downto 0);
					ideoutd(14 downto 10) <= addonin(8 downto 4);
					ideoutd(15) <= ideoutdd(10);
				ELSE
					lrclk<=ideoutdd(1);
					adat<=ideoutdd(0);
				END IF;
				IF sync='1' THEN
					ms_datai <= addonin(1);
					ms_clki <= addonin(2);
				ELSE
					kb_datai <= addonin(1);
					kb_clki <= addonin(2);
				END IF;
			END IF;
			IF Ring(1)='1' THEN
				ide_wrd <= ide_wrin;
				ide_rd <= addonin(1);
				ide_wrin <= addonin(2);

				idea <= addonin(6 downto 4);
				ideoutdd(10) <= addonin(8);
			END IF;
			IF Ring(2)='1' THEN
				
				ideoutdd(4 downto 0) <= addonin(8 downto 4);
			END IF;
			IF Ring(3)='1' THEN
				sync <=  inbuf(2);
				TxD <= addonin(0);
				ide_csp0 <= addonin(1);
--		ide_csp0 <= NOT addonin(2);
				ide_css0 <= addonin(2);
				ideoutdd(9 downto 5) <= addonin(8 downto 4);
			END IF;
		END IF;
	END PROCESS; 
-------------------------------------------------------------------	
--  TRANSMITTER		
--			0		1		2			3     

--addonout
--(0)		idei(0)	idei(5)	idei(10)	idei(15)	
--(1)		idei(1)	idei(6)	idei(11)		
--(2)		idei(2)	idei(7)	idei(12)	kb_clk
--(3)		idei(3)	idei(8)	idei(13)	kb_data
--(4)		idei(4)	idei(9)	idei(14)	ms_clk
--(5)		C2MHz	ioshift				ms_data
	PROCESS(addonin)
	BEGIN
		IF addonin(3)'event AND addonin(3) = '1' THEN
			C2MHzd <= C2MHz;
			IF Ring(0)='1' THEN
				addonout(4 downto 0) <= idedata(4 downto 0);
--				addonout(4 downto 0) <= "10100";
				addonout(5) <= C2MHzd;
				ideind <= idedata(15 downto 5);
--				ideind <= "10101001010";
			END IF;
			IF Ring(1)='1' THEN
				addonout(4 downto 0) <= ideind(4 downto 0);
				addonout(5) <= ioshift;
			END IF;
			IF Ring(2)='1' THEN
				addonout(4 downto 0) <= ideind(9 downto 5);
--				addonout(5) <= ide_csp0;
--				addonout(5) <= sync;
			END IF;
			IF Ring(3)='1' THEN
				addonout(0) <= ideind(10);
				addonout(2) <= kb_clk;
				addonout(3) <= kb_data;
				addonout(4) <= ms_clk;
				addonout(5) <= ms_data;
			END IF;
		END IF;
	END PROCESS; 
END;  
