	.include "drivecpu.i"

	.export fat_findfile
	.export fat_loadfile
	.export fat_cdroot
	.export fat_next_clust
	.export fat_read_next_sector

	.exportzp part_cstart
	.exportzp dir_c1dcore	; 32-bit start of c1dcore
	.exportzp dir_core		; 32-bit start of core
	.exportzp vol_secperclus
	.exportzp part_rootdir
	.exportzp vol_fstype
	.exportzp scount
	.exportzp part_num
	.exportzp out_page
	.exportzp outptr
	.exportzp mem_page
	.exportzp sectorptr
	.exportzp tmp_page
	.exportzp tmpptr
	.exportzp file_length
	.exportzp cluster
	
	.import ide_read_sector
	.import ide_read_sector_2buf
	.import mkicon
	
	.exportzp nameptr
	.exportzp dir_deep
	.exportzp lba

	.import debug_puts
	.import debug_puthex
	.import debug_putdigit
	.import debug_put
	.import debug_crlf
	
	.zeropage
	
part_fat:		.res 4	; 32-bit start of fat
part_cstart:	.res 4	; 32-bit start of clusters
part_rootdir:	.res 4	; 32-bit root directory cluster or
						; start sector for FAT12/FAT16 root directory
part_rootdirsize:	.res 1	; size of FAT12/FAT16 root directory
vol_fstype:		.res 1	
vol_secperclus:	.res 1
part_num:		.res 1	; device/partition number

lba:			.res 4	; 32-bit block address
tmp_page:		.res 1	; use with lamy/samy
tmpptr:			.res 2	; pointer to where data is loaded
;-----------------------------------------------------
out_page:		.res 1	; use with lamy/samy
outptr:			.res 2	; pointer to where data is loaded
mem_page:		.res 1	; use with lamy/samy
sectorptr:		.res 2	; pointer to where data is loaded
dir_deep:		.res 1
dir_c1dcore:	.res 4	; 32-bit start of c1dcore
dir_core:		.res 4	; 32-bit start of core
nameptr:		.res 2	; pointer to where data is loaded
cluster:	.res 4	; 32-bit start of clusters
scount:		.res 1	; number of sectors to read
file_length:	.res 4	; 32-bit start of clusters
fat12tmp:	.res 2	; work area to compute fat 12 sectors
fat12half:	.res 1	; flag that tells us if we should start with the
			; high nybble ($80) or low nybble ($00)
fat12byte:	.res 1	; temporary storage for current fat 12 byte
fat12cluster:	.res 1  ; more temporary storage
fat12fatflag:	.res 1	; flag that tells us to read the next fat sector

	.code
	jmp	ide_read_sector_2buf	;$f800
	jmp	fat_cdroot				;$f803
	jmp	fat_read_next_sector	;$f806
	jmp	fat_findfile			;$f809
	jmp	fat_loadfile			;$f80c
	jmp mkicon					;$f80f
	
fat_findfile:
	stax nameptr
	jsr ide_read_sector_2buf	; load dir sector into buffer
	bcs @error
@checksector:	
	ldax #sectorbuf	; load data into sectorbuf
	stax sectorptr
	
@checkentry:
	ldy #11			; check entry type
	lda (sectorptr),y
	and #$18
	bne @skip
	
	ldy #0			
	lda (sectorptr),y
	beq @lastentry
	cmp #$e5		;deleted
	beq @skip
	
@next_char:	
	lda (nameptr),y
	cmp #'?'
	beq	:+
	sec
	sbc (sectorptr),y
	beq	:+
	cmp #$e0
	bne @skip
:	
	iny
	cpy #11;namelen
	bne @next_char
	beq @get_file

@skip:	
	lda sectorptr
	clc
	adc #$20
	sta sectorptr
	bcc @checkentry
	inc sectorptr+1
	lda sectorptr+1
	and #1				;stop at $04
	bne @checkentry
	
	lda	scount
	cmp #1
	bne	:+
	
:	jsr fat_read_next_sector
	bcc @checksector		; premature end of dir
	
@lastentry:

;	rts
@error:
	sec
	rts

@get_file:	

;	lda #'F'
;	sta $b000
;; copy start cluster and length from dirptr entry
;fat12_stat:
;fat16_stat:
	lda #0			; upper half is 0
	sta cluster+2
	sta cluster+3
;	sta dir_is_fat16root

	lda vol_fstype
	cmp #fs_fat32
	bne :+
;fat32_stat:
	ldy #$14		; copy the start cluster
	lda (sectorptr),y
	sta cluster+2
	iny
	lda (sectorptr),y
	sta cluster+3

:
	ldy #$1a
	lda (sectorptr),y
	sta cluster
	iny
	lda (sectorptr),y
	sta cluster+1
	
	ldy #$1f		; copy the length
	ldx #3
:	lda (sectorptr),y
	sta file_length,x
	dey
	dex
	bpl :-
	
	clc
	rts

fat_loadfile:
@next_cluster:
	jsr clustertolba	; calculate lba address, sets scount
@done:	
	lda out_page
	sta mem_page
	ldax outptr
	stax sectorptr
	jsr ide_read_sector	; read the sector
;		lda #'-'
;		bcs :+
;		lda #'*'
;:		sta $800
	lda out_page
	bne	:+
	lda outptr+1 
	and #$f0	;IO page 0
	beq @skipptr
	
:	lda mem_page
	sta out_page
	ldax sectorptr
	stax outptr
@skipptr:
;test length
	sec
	lda	file_length+1
	sbc	#2
	sta	file_length+1
	bne @subsect
	ora file_length
	ora file_length+2
	ora file_length+3
	bne @scount
@complete:
	lda	#0
	sec			; end of file
	rts	
	
@subsect:
	lda file_length+2
	sbc #0
	sta file_length+2
	lda file_length+3
	sbc #0
	sta file_length+3
	bcc @complete
	
@scount:
	dec scount
	beq :+ 
	
;next_sector:
	inc lba			; next sector
	bne @done
	inc lba+1
	bne @done
	inc lba+2
	bne @done
	inc lba+3
	bne @done
;-----------------	
:
	jsr	fat_next_clust
	bne	@next_cluster
	sec			; end of cluster chain
	rts	


	.code

fat12_next_byte:
	ldx fat12tmp
	lda fat12tmp+1
	bne @upperhalf

	lda sectorbuf,x		; get byte from lower half of FAT sector
	jmp @getnext

@upperhalf:
	lda sectorbuf + 256,x	; get byte from upper half of FAT sector

@getnext:
	sta fat12byte
	inc fat12tmp		; increment our byte offset to the FAT
	bne @done
	inc fat12tmp+1
	lda fat12tmp+1
	cmp #$02		; check if we're past the end of the sector
	bcc @done
	
	bit fat12fatflag	; should we read the next fat sector ?
	bmi @done
	
	lda #$00		; zero out or fat sector offset
	sta fat12tmp+1
	
	inc lba			; read in the next sector
	bne @skiplb1
	inc lba+1
@skiplb1:

	jsr ide_read_sector_2buf
	bcs @error

@done:
	lda fat12byte
	clc
	rts

@error:
	sec
	rts


fat12_next_clust:
				; We have to multiply the cluster number
				; Times 1.5 to get a byte offset into the FAT
	clc
	lda cluster+1
	lsr			; Divide by two first
	sta fat12tmp+1
	lda cluster
	ror
	sta fat12tmp
	lda #$00
	ror
	sta fat12half		; Set our flag to know if we have to start
				; on the high nybble
	;clc
	lda cluster		; Add the result of the division by two to
	adc fat12tmp		; the original cluster number and we get
	sta fat12tmp		; cluster*1.5 + our 'half' flag
	lda cluster+1
	adc fat12tmp+1
	sta fat12tmp+1		; We now have the full byte offset in fat12tmp
	
	lsr			; bits 15-9 contain our fat sector number
	sta lba			; so shift them in place
	lda #$00
	sta lba+1
	sta lba+2
	sta lba+3
	
	lda fat12tmp+1		; Strip the highest 7 bits and we get a
	and #$01		; byte offset into the sector
	sta fat12tmp+1			
	
	jsr addfatstart

	jsr ide_read_sector_2buf
	bcs @error

@iscached:
	lda #$7f		; set flag to read the next sector if
	sta fat12fatflag	; necessary
	jsr fat12_next_byte	; Get the current byte from the FAT
	bcs @error
	sta fat12cluster
	
	inc fat12fatflag	; don't read the next sector this time
	jsr fat12_next_byte	; Next byte
	bcs @error

	bit fat12half		; test nybble flag
	bmi @nybswap
	and #$0f		; Flag not set so just mask out the 4
	sta cluster+1		; high bits of the second byte
	lda fat12cluster
	sta cluster
	jmp @checklast
	
@nybswap:
	lsr
	ror fat12cluster	; Flag set, so rotate things into place
	lsr
	ror fat12cluster
	lsr
	ror fat12cluster
	lsr
	ror fat12cluster
	sta cluster+1
	lda fat12cluster
	sta cluster
	
@checklast:
	lda cluster+1
	cmp #$0f
	bne @done
	lda cluster
	and #$f0
	cmp #$f0
	
@done:
	clc
	rts

@error:
	sec
	rts


fat_next_clust:
	lda vol_fstype
	cmp #fs_fat12
	bne @testfat32
	jmp fat12_next_clust	; We've got range
@testfat32:
	cmp #fs_fat32
	beq fat32_next_clust


; find the next linked cluster from the FAT
fat16_next_clust:
	lda cluster+1		; there are 2 bytes for each entry
	sta lba			; in the FAT this gives us 256
	lda cluster+2		; entries in every FAT sector so
	sta lba+1		; bits 31..8 of the current cluster
	lda cluster+3		; address gives us the sector to
	sta lba+2		; read
	lda #0
	sta lba+3

	jsr addfatstart		; fat += part-fat

	jsr ide_read_sector_2buf
	bcs @error

@iscached:
	lda cluster		; offset = cluster<<1
	asl
	tax
	bcs @upperhalf

	lda sectorbuf,x		; copy new cluster address
	sta cluster
	inx
	lda sectorbuf,x
	sta cluster+1

	jmp @checkeoc

@upperhalf:
	lda sectorbuf + 256,x
	sta cluster
	inx
	lda sectorbuf + 256,x
	sta cluster+1

@checkeoc:
	lda #0		; this is FAT*16*
	sta cluster+2
	sta cluster+3

	lda cluster
	ora	#$0f
	and cluster+1		; check for end of cluster chain
	cmp #$ff
;	bne @done

;	lda cluster
;	and #$f0		; the 4 least significant bits are ignored
;	cmp #$f0

@done:
	clc
	rts

@error:
	sec
	rts


fat32_next_clust:
	lda cluster+1		; there are 4 bytes for each entry
	sta lba			; in the FAT this gives us 128 entries
	lda cluster+2		; in every FAT sector so bits 31..7 of
	sta lba+1		; the current cluster address gives us
	lda cluster+3		; the FAT sector to read
	sta lba+2
	lda #0
	sta lba+3

	lda cluster		; lba = cluster>>7
	asl
	rol lba
	rol lba+1
	rol lba+2
	rol lba+3
	
	jsr addfatstart		; fat += part_fat
	
	jsr ide_read_sector_2buf
	bcs @error

@iscached:
	lda cluster		; offset = (cluster & 127)<<2
	asl
	asl
	tax
	bcs @upperhalf

	lda sectorbuf,x		; copy new cluster address
	sta cluster
	inx
	lda sectorbuf,x
	sta cluster+1
	inx
	lda sectorbuf,x
	sta cluster+2
	inx
	lda sectorbuf,x
	sta cluster+3

	jmp @checkeoc

@upperhalf:
	lda sectorbuf + 256,x
	sta cluster
	inx
	lda sectorbuf + 256,x
	sta cluster+1
	inx
	lda sectorbuf + 256,x
	sta cluster+2
	inx
	lda sectorbuf + 256,x
	sta cluster+3

@checkeoc:
	lda cluster+3		; check for end of cluster chain
	ora	#$f0
	and cluster+2
	and cluster+1
	cmp #$ff
	bne @done

	lda cluster
;	and #$f0		; the 4 least significant bits are ignored
;	cmp #$f0
	and #$f8		; the 3 least significant bits are ignored
	cmp #$f8

@done:
	clc
	rts

@error:
	sec
	rts


; lba += start of fat
addfatstart:
	clc
	lda lba
	adc part_fat
	sta lba
	lda lba+1
	adc part_fat+1
	sta lba+1
	lda lba+2
	adc part_fat+2
	sta lba+2
	lda lba+3
	adc part_fat+3
	sta lba+3
	rts

; load the specified cluster to memory
fat_read_next_sector:
	dec scount
	beq @next_cluster
	
;fat_read_next_sector:
	inc lba			; next sector
	bne @done
	inc lba+1
	bne @done
	inc lba+2
	bne @done
	inc lba+3
@done:	
	jmp ide_read_sector_2buf	; read the sector
	
@next_cluster:	
	jsr	fat_next_clust
	beq	@end
	jsr	clustertolba
	jmp ide_read_sector_2buf	; read the sector

@end:
	sec			; end of cluster chain
	rts	

	
; load the root directory
fat_cdroot:
	lda dir_deep
	bne clustertolba3
	
	lda vol_fstype
	cmp #fs_fat32
	beq clustertolba2
	
; calculate lba address of cluster
; with ugly, ugly special case for FAT12/16 root directory
; sets scount

	lda #0
	sta cluster
	sta cluster+1
	sta cluster+2
	sta cluster+3
	
	lda part_rootdir	; fat12/16 
	sta lba
	lda part_rootdir+1
	sta lba+1
	lda part_rootdir+2
	sta lba+2
	lda part_rootdir+3
	sta lba+3
	lda part_rootdirsize	; assuming rootdir entries for FAT16 
	sta scount
	rts
	
clustertolba2:
	lda part_rootdir	; lba = cluster * secperclus + part_cstart
	sta cluster
	lda part_rootdir+1
	sta cluster+1
	lda part_rootdir+2
	sta cluster+2
	lda part_rootdir+3
	jmp clustertolba4

clustertolba3:
	and #1
	beq clustertolba5
	lda dir_c1dcore	; lba = cluster * secperclus + part_cstart
	sta cluster
	lda dir_c1dcore+1
	sta cluster+1
	lda dir_c1dcore+2
	sta cluster+2
	lda dir_c1dcore+3
	jmp clustertolba4

clustertolba5:
	lda dir_core	; lba = cluster * secperclus + part_cstart
	sta cluster
	lda dir_core+1
	sta cluster+1
	lda dir_core+2
	sta cluster+2
	lda dir_core+3
clustertolba4:	
	sta cluster+3

clustertolba:
	lda cluster	; lba = cluster * secperclus + part_cstart
	sta lba
	lda cluster+1
	sta lba+1
	lda cluster+2
	sta lba+2
	lda cluster+3
	sta lba+3

	lda vol_secperclus	; multiply cluster address by number of
	sta scount
@shift:
	lsr			; sectors per cluster. this number is a
	bcs @shifted		; power of two so we'll just shift.
	asl lba
	rol lba+1
	rol lba+2
	rol lba+3
	jmp @shift

@shifted:
	clc			; add start of clusters
	lda lba
	adc part_cstart
	sta lba
	lda lba+1
	adc part_cstart+1
	sta lba+1
	lda lba+2
	adc part_cstart+2
	sta lba+2
	lda lba+3
	adc part_cstart+3
	sta lba+3
	; we should now have the correct cluster in the lba address
	rts

