	.include "ide.i"
	.include "drivecpu.i"
 
	.importzp lba
	.importzp mem_page
	.importzp sectorptr
	.importzp part_num

	.export ide_init
	.export ide_read_sector
	.export ide_read_sector_2buf
	.export ide_read_status
	.export ide_read_error

	.export atapi_init

	.export ide_channel
	.export ide_device
	.export ide_drivetype
	.export idecache_valid	

	.exportzp devtype_none
	.exportzp devtype_hd
	.exportzp devtype_cd

	.zeropage

ide_channel:		.res 1	; primary ($00) or secondary ($08)
ide_device:			.res 1	; master ($00) or slave ($10)
pagecount:		.res 1	; keeps track of how much data to load
wait_drq_timeout:	.res 1	; timeout counter
wait_ready_timeout:	.res 1
init_timeout:		.res 1
ide_drivetype:		.res 1

idecache:			.res 4	; 32-bit currently cached ide sector
idecache_valid:		.res 1 

devtype_none	= $00
devtype_hd	= $01
devtype_cd	= $02
;
;
;	.rodata
	.code

devtab:
	.byte $E0	; primary master
	.byte $F0	; primary slave
	.byte $E8	; secondary master
	.byte $F8	; secondary slave
	.code
; initialize device. call with dev 0..3 in A
atapi_init:
ide_init:
	sta part_num
	
	tax
;set sectorcache to invalid	
	lda #$55
	sta idecache_valid
	
	lda devtab,x
	and #$0f
	sta ide_channel
	eor devtab,x
	sta ide_device
	ldy #ide_lba3
	jsr ide_write_reg

	jsr delay_400ns

	jsr ide_read_status
	beq @done

@checkrdy:
	lda #0
	sta init_timeout

:	jsr ide_wait_ready	; wait for ready status, with timeout
	bcc @done

	jsr delay_400ns
	inc init_timeout
	bne :-

	jsr ide_read_status
	lda #devtype_none
	sta ide_drivetype
	sec
	rts
	
@done:
	
	lda #idecmd_identify		; identify HD
	ldy #ide_command
	jsr ide_write_reg		; we don't want to get stuck here...

	jsr delay_400ns
	
	jsr ide_read_error		; check for error condition	
	bcc @foundhd

	lda #idecmd_identifypacket	; identify CD
	ldy #ide_command
	jsr ide_write_reg		; we don't want to get stuck here...
	jsr delay_400ns
	jsr ide_read_error		; check for error condition
	bcs @failed			; ok, no CD

	lda #devtype_cd			; set type to CD
	sta ide_drivetype

	sec
	rts

@foundhd:
	lda #devtype_hd			; set type to HD
	sta ide_drivetype

@readdata:

	jsr ide_wait_drq		; wait for data if BSY drops and DRQ
	bcs @failed			; is not set, we're fux0red

	lda	#0
	sta mem_page
	ldax #sectorbuf	; load data into sectorbuf
	stax sectorptr

	jsr ide_read_256_words
	bcs @failed

	lda ide_drivetype
	clc
	rts

@failed:
	lda #devtype_none
	sta ide_drivetype
	sec
	rts


; read 512-byte sector to sectorptr
ide_read_sector_2buf:
; check if sector is already cached
	lda idecache_valid
	cmp	#$aa
	bne @no
	
	ldx #3
:	lda lba,x
	cmp idecache,x
	bne @no
	dex
	bpl :-
; sector alway loaded	
	clc
	rts
	
@no:
	lda #0
	sta mem_page
	ldax #sectorbuf	; load data into sectorbuf
	stax sectorptr
	
	jsr ide_read_sector
	lda #$55
	bcs @exit
	
	ldx #3
:	lda lba,x
	sta idecache,x
	dex
	bpl :-
	lda #$aa
@exit:	
	sta idecache_valid
	rts
	
ide_read_sector:
	jsr setlba		; set lba registers

	ldy #ide_scount		; read a single sector
	lda #1
	jsr ide_write_reg

	ldy #ide_command	; send ide read command
	lda #idecmd_read_sector
	jsr ide_write_reg

	jsr delay_400ns		; delay after sending command

	jsr ide_read_error
	bcc ide_read_256_words
	rts


ide_read_256_words:
	jsr ide_wait_drq	; wait for DRQ
	bcs @timeout

	lda #2			; number of pages to load
	sta pagecount

	ldy #0			; store data at sectorptr

	lda #ide_status
	ora ide_channel
	tax
@checkstatus:
	lda	ide_base,x
	and #$08		; check that DRQ is set
	beq @checkstatus
	ldx ide_channel
;	tax
@next:
	lda	ide_base,x
	samy mem_page
	
	iny
	lda	ide_base,x
	samy mem_page
	
	iny
	bne @next
	inc sectorptr+1	;tobix
	bne @nextp
	inc mem_page
@nextp:	
	dec pagecount
	bne @next

	clc
@timeout:
	rts

;; set LBA address
setlba:
	ldy #ide_lba3
	lda lba+3
	and #$0f
	ora ide_device		; lba addressing
	jsr ide_write_reg

	jsr delay_400ns		; delay after selecting device

	ldy #ide_lba2
	lda lba+2
	jsr ide_write_reg

	ldy #ide_lba1
	lda lba+1
	jsr ide_write_reg

	ldy #ide_lba0
	lda lba
	jmp ide_write_reg


; wait for DRQ, with timeout
ide_wait_drq:
	lda #ide_status
	ora ide_channel
	tax
	lda #0
	sta wait_drq_timeout
@checkstatus:
	lda	ide_base,x
	and #$88		; check for BSY not set
	bmi @checkstatus
	and #$08		; check that DRQ is set
	bne @ready

	inc wait_drq_timeout
	bne @checkstatus

	sec
	rts

@ready:	clc
	rts


; wait for RDY
ide_wait_ready:
	lda #0
	sta wait_ready_timeout

	lda #ide_status
	ora ide_channel
	tax
@checkstatus:
	lda	ide_base,x
	and #$c0
	bmi @checkstatus
	and #$40
	bne @done

	inc wait_ready_timeout
	bne @checkstatus

	sec
	rts

@done:	clc
	rts


; write A to register in Y
ide_write_reg:
	cpy #7
	beq @write_command
	pha
	lda #ide_status		; read ide status
	ora ide_channel
	tax
@checkstatus:	
	lda	ide_base,x
	and #$80		; check that BSY isn't set
	bne @checkstatus
	tya			; write to register in Y
	ora ide_channel
	tax
	pla
	sta	ide_base,x	; write A
	rts

@write_command:
	pha
	cmp #idecmd_identify
	beq @do_write
	cmp #idecmd_identifypacket
	beq @do_write
	lda #ide_status		; read ide status
	ora ide_channel
	tax
:	lda	ide_base,x
	eor #$40		; check that RDY is set and
	and #$c0		; that BSY isn't set
	bne :-
@do_write:
	tya			; write to register in Y
	ora ide_channel
	tax
	pla
	sta	ide_base,x	; write A
	rts

; read the IDE status register
ide_read_status:
	lda #ide_status		; read ide status
	ora ide_channel
	tax
	lda	ide_base,x
	cmp #$00
	rts


; read the error register, return error status in carry
ide_read_error:
	lda #ide_status
	ora ide_channel
	tax
@read:	
	lda	ide_base,x
	and #$81
	bmi @read
	lsr
	lda #ide_error
	ora ide_channel
	tax
	lda	ide_base,x
	rts

; 400 ns delay
delay_400ns:
	ldx #10 * cpuspeed
:	nop
	bit $ea
	nop
	bit $ea
	nop
	bit $ea
	nop
	bit $ea
	nop
	bit $ea
	nop
	bit $ea
	nop
	bit $ea
	dex
	bne :-			; 40*x - 1
	rts
