------------------------------------------------------------------------------
------------------------------------------------------------------------------
--                                                                          --
-- Copyright (c) 2006-2009 Tobias Gubener                                   -- 
-- Subdesign CPC T-REX by TobiFlex                                          --
--                                                                          --
-- This source file is free software: you can redistribute it and/or modify --
-- it under the terms of the GNU General Public License as published        --
-- by the Free Software Foundation, either version 3 of the License, or     --
-- (at your option) any later version.                                      --
--                                                                          --
-- This source file is distributed in the hope that it will be useful,      --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of           --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            --
-- GNU General Public License for more details.                             --
--                                                                          --
-- You should have received a copy of the GNU General Public License        --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.    --
--                                                                          --
------------------------------------------------------------------------------
------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity sdram is
port
	(
	zena_in		: in std_logic;
	sdata		: inout std_logic_vector(15 downto 0);
	sdaddr		: out std_logic_vector(11 downto 0);
	dqm			: out std_logic_vector(1 downto 0);
	ba			: buffer std_logic_vector(1 downto 0);
	sd_we		: out std_logic;
	sd_ras		: out std_logic;
	sd_cas		: out std_logic;

	sysclk		: in std_logic;
	reset		: in std_logic;
	
	zdatawr		: in std_logic_vector(7 downto 0);
	zAddr		: in std_logic_vector(23 downto 0);
	zwr			: in std_logic;
	wsgin			: in std_logic;			-- '0'-cache off '1'-wsg/cache on
	zstate		: in std_logic_vector(1 downto 0);
	
	vAddr		: in std_logic_vector(23 downto 0);
	vrefresh	: in std_logic;
	
	datawr		: in std_logic_vector(7 downto 0);
	rAddr		: in std_logic_vector(23 downto 0);
	rmem		: in std_logic;
	rwr			: in std_logic;
	rrefresh	: in std_logic;
	rstate		: in std_logic_vector(1 downto 0);
	speed		: in std_logic;			-- '0'-Clone '1'-Turbo 
	mdb			: in std_logic_vector(7 downto 0);
	resetin		: in std_logic;
	shresetin		: in std_logic;
	
	dataout		: out std_logic_vector(7 downto 0);
	pixelout	: out std_logic_vector(7 downto 0);
	zdataout		: out std_logic_vector(7 downto 0);
	cpuclk		: out std_logic;
	rena_o		: out std_logic;
	zena_o		: out std_logic;
	zwait		: out std_logic;
	vena		: out std_logic;
	reset_out	: buffer std_logic;
	c1MHz		: out std_logic;
	ring		: out std_logic_vector(3 downto 0);
	IOB_clk 	: out std_logic;
	cpcreset 	: buffer std_logic;
	shreset 	: buffer std_logic;
	wsg			: buffer std_logic
	);
end;

architecture rtl of sdram is


signal initstate	:std_logic_vector(3 downto 0);
signal slow			:std_logic_vector(7 downto 0);
signal video		:std_logic_vector(63 downto 0);
signal datard		:std_logic;
signal sdcom_ena	:std_logic;
signal isd_ras		:std_logic;
signal isd_cas		:std_logic;
signal isd_we 		:std_logic;
signal wr_int		:std_logic;
signal rd_int		:std_logic;
signal init_done	:std_logic;
signal Addr			:std_logic_vector(23 downto 0);
signal nextpix		:std_logic_vector(15 downto 0);
signal nextzdata	:std_logic_vector(15 downto 0);
signal isdaddr		:std_logic_vector(11 downto 0);
signal datain		:std_logic_vector(15 downto 0);
signal datamux		:std_logic_vector(15 downto 0);
signal casaddr		:std_logic_vector(23 downto 0);
signal Latency 		:std_logic;
signal wr 			:std_logic;
signal refresh 		:std_logic;
signal sdwrite 		:std_logic;
signal rcache		:std_logic_vector(63 downto 0);
signal rcache_addr	:std_logic_vector(23 downto 0);
signal fill_rcache	:std_logic;

signal zcache		:std_logic_vector(63 downto 0);
signal zcache_addr	:std_logic_vector(23 downto 0);
signal zcache_recall	:std_logic;
signal fill_zcache	:std_logic;

signal rena			:std_logic;
signal rvalid		:std_logic_vector(3 downto 0);
signal zena			:std_logic;
signal V_cycle		:std_logic;
signal R_cycle		:std_logic;
signal Z_cycle		:std_logic;
signal dwrL		:std_logic;
signal dwrU		:std_logic;
signal zvalid		:std_logic_vector(3 downto 0);
signal rmwait		:std_logic;
signal zmwait		:std_logic;
signal requal		:std_logic;
signal zequal		:std_logic;
signal zena_clone	:std_logic;

signal S_cycle		:std_logic;
signal wr_addr		: std_logic_vector(23 downto 0);
signal wrd			: std_logic;
signal dma			: std_logic;
signal dmad			: std_logic;
signal mdbd: std_logic_vector(7 downto 0);
signal mdcdd: std_logic_vector(7 downto 0);
signal shift: std_logic_vector(8 downto 0);
signal romready		: std_logic;
signal wrd_valid	: std_logic;

type sdram_states is (ph0,ph1,ph2,ph3,ph4,ph5,ph6,ph7);
signal sdram_state		: sdram_states;
type pass_states is (nop,ras,cas);
signal pass		: pass_states;

begin

-------------------------------------------------------------------------
-- Shadow cache
-------------------------------------------------------------------------
	rena_o <= '1' when ((rena='1' and wrd_valid='1') or rmem='1' or rmwait='1' or rstate="01") else '0'; 
	
	process (sysclk, rcache, rvalid, requal, rstate, raddr, rcache_addr, nextpix) 
	begin
		if raddr(23 downto 3)=rcache_addr(23 downto 3) then
			requal <='1';
		else	
			requal <='0';
		end if;	
		
		rmwait <= '0';
		if requal='1' and rvalid(0)='1' then
			case (raddr(2 downto 1)-rcache_addr(2 downto 1)) is
				when "00"=>
					rmwait <= rvalid(0);
					if raddr(0)='1' then
						dataout <= rcache(55 downto 48);
					else
						dataout <= rcache(63 downto 56);
					end if;
				when "01"=>
					rmwait <= rvalid(1);
					if raddr(0)='1' then
						dataout <= rcache(39 downto 32);
					else
						dataout <= rcache(47 downto 40);
					end if;
				when "10"=>
					rmwait <= rvalid(2);
					if raddr(0)='1' then
						dataout <= rcache(23 downto 16);
					else
						dataout <= rcache(31 downto 24);
					end if;
				when "11"=>
					rmwait <= rvalid(3);
					if raddr(0)='1' then
						dataout <= rcache(7 downto 0);
					else
						dataout <= rcache(15 downto 8);
					end if;
				when others=> null;
			end case;	
		else	
			if raddr(0)='1' then
				dataout <= nextpix(7 downto 0);
			else
				dataout <= nextpix(15 downto 8);
			end if;
		end if;	
	end process;		
	
--Datenbernahme
	process (reset, sysclk) begin
		if reset = '0' then
			fill_rcache <= '0';
			rena <= '0';
			rvalid <= "0000";
		elsif (sysclk'event and sysclk='1') then
				if requal='1' and rwr='0' then
					rvalid <= "0000";
				end if;
					case sdram_state is	
						when ph0 =>	
										if fill_rcache='1' then
											rcache(47 downto 32) <= nextpix;
										end if;
						when ph1 =>	
										if fill_rcache='1' then
											rvalid(1) <= '1';
											rvalid(2) <= '1';
											rcache(31 downto 16) <= nextpix;
										end if;
						when ph2 =>	
										if fill_rcache='1' then
											rcache(15 downto 0) <= nextpix;
										end if;
						when ph3 =>	
										if fill_rcache='1' then
											rvalid(3) <= '1';
										end if;
										fill_rcache <= '0';
--						when ph4 =>	
						when ph5 =>
										rena <= R_cycle;
										IF wrd_valid='1' AND R_cycle='1' THEN
											if rstate="00" then
												rcache_addr <= raddr;
												fill_rcache <= '1';
												rvalid <= "0000";
											end if;
										END IF;
--						when ph6 =>	
						when ph7 =>	
										rena <= '0';
										if fill_rcache='1' then
											rcache(63 downto 48) <= nextpix;
											rvalid(0) <= '1';
										end if;
						when others => null;
					end case;	
			end if;
	end process;	
		
-------------------------------------------------------------------------
-- CPC cache
-------------------------------------------------------------------------
	zena_o <= '1' when (zena='1' and wrd_valid='1') or zena_clone='1' or ((zmwait='1' or zstate="01") and wsg='1') else '0'; 
--	zwait <= Z_cycle;
	process (sysclk, zaddr, zcache_addr, zvalid, zcache, wsg, zstate, zequal, nextzdata) begin
		if zaddr(23 downto 3)=zcache_addr(23 downto 3) then
			zequal <='1';
		else	
			zequal <='0';
		end if;	
		zmwait <= '0';
		if zequal='1' and zvalid(0)='1' then
			case (zaddr(2 downto 1)-zcache_addr(2 downto 1)) is
				when "00"=>
					zmwait <= zvalid(0);
					if zaddr(0)='1' then
						zdataout <= zcache(55 downto 48);
					else
						zdataout <= zcache(63 downto 56);
					end if;
				when "01"=>
					zmwait <= zvalid(1);
					if zaddr(0)='1' then
						zdataout <= zcache(39 downto 32);
					else
						zdataout <= zcache(47 downto 40);
					end if;
				when "10"=>
					zmwait <= zvalid(2);
					if zaddr(0)='1' then
						zdataout <= zcache(23 downto 16);
					else
						zdataout <= zcache(31 downto 24);
					end if;
				when "11"=>
					zmwait <= zvalid(3);
					if zaddr(0)='1' then
						zdataout <= zcache(7 downto 0);
					else
						zdataout <= zcache(15 downto 8);
					end if;
				when others=> null;
			end case;	
		else	
			if zaddr(0)='1' then
				zdataout <= nextzdata(7 downto 0);
			else
				zdataout <= nextzdata(15 downto 8);
			end if;
		end if;	
	end process;		
		
	
	
--Datenbernahme
	process (reset, sysclk) begin
		if reset = '0' then
			fill_zcache <= '0';
			zena <= '0';
			zvalid <= "0000";
		elsif (sysclk'event and sysclk='1') then
				if zequal='1' and zstate="11" then
					zvalid <= "0000";
				end if;
					case sdram_state is	
						when ph0 =>	
										if fill_zcache='1' then
											zcache(47 downto 32) <= nextpix;
										end if;
						when ph1 =>		
										if fill_zcache='1' then
											zvalid(1) <= '1';
											zvalid(2) <= '1';
											zcache(31 downto 16) <= nextpix;
										end if;
						when ph2 =>	
										if fill_zcache='1' then
											zcache(15 downto 0) <= nextpix;
										end if;
						when ph3 =>		if zena_clone='0' then
											wsg <= wsgin;
										end if;
										if fill_zcache='1' then
											zvalid(3) <= '1';
										end if;
										fill_zcache <= '0';
						when ph4 =>		

						when ph5 =>		zena <= Z_cycle;
										zwait <= Z_cycle;
										if wrd_valid='1' AND Z_cycle='1' then
											if zstate="00" then
												zcache_addr <= zaddr;
												fill_zcache <= '1';
												zvalid <= "0000";
											end if;
										end if;
										if wsg='1' then
											zena_clone <= '0';
										else	
											zena_clone <= Z_cycle;
											zvalid <= "0000";
										end if;
						when ph6 =>	
										IF Z_cycle='1' THEN
											nextzdata <= sdata;
										END IF;
						when ph7 =>	
										zena <= '0';
										zwait <= '0'; 
										if fill_zcache='1' then
											zcache(63 downto 48) <= nextpix;
											zvalid(0) <= '1';
										end if;
						when others =>
					end case;	
			end if;
	end process;		
	
-------------------------------------------------------------------------
-- Access slot control
-------------------------------------------------------------------------
	vena <= V_cycle;
	process (sysclk) begin
			if (sysclk'event and sysclk='1') then
				cpcreset <= cpcreset AND resetin AND reset;
				shreset <= shreset AND shresetin AND reset;
				nextpix <= sdata;
				
				if dma='0' then
					wr_addr <= X"090000";
				end if;
				if sdram_state=ph7 then
					if V_cycle='1' then
						if casaddr(0)='1' then
							pixelout <= nextpix(7 downto 0);
						else
							pixelout <= nextpix(15 downto 8);
						end if;
					end if;
					if S_cycle='1' THEN
						wr_addr <= wr_addr+1;
					end if;	
					R_cycle <= '0';
					S_cycle <= '0';
					V_cycle <= '0';
					Z_cycle <= '0';
					IF slow(2 downto 0)="101" THEN
						c1MHz <= NOT slow(3);
						IF slow(3)='0' THEN
							cpcreset <= resetin;
							shreset <= shresetin;
						END IF;
					END IF;
				end if;	
				
				if sdram_state=ph0 then
					IF slow(2 downto 0)=5 THEN
						slow <= slow+3;
					ELSE
						slow <= slow+1;
					END IF;
					wrd <= shift(0);

					IF wrd='1' AND shift(0)='0' AND romready='0' THEN  
						S_cycle <= '1';		--store init
					ELSE	
						IF wsg='1' THEN
							IF slow(2 downto 0)="100" THEN
								V_cycle <= '1';
							ELSIF zena_in='0' THEN
								Z_cycle <= '1';
							ELSE
								R_cycle <= '1';
							END IF; 	  
						ELSIF speed='1' THEN			--Turbo
							case slow(3 downto 0) is
								when "0100"|"1100" => V_cycle <= '1';
								when "0010"|"0000"|"1010"|"1000" => 
									R_cycle <= '1';
								when others =>
									IF zena_in='0' THEN
										Z_cycle <= '1';
									ELSE
										R_cycle <= '1';
									END IF;
	 							end case;
						ELSE						--Clone
							case slow(3 downto 0) is
								when "0100"|"1100" => V_cycle <= '1';
								when "0010" => Z_cycle <= '1';
								when others => R_cycle <= '1';
							end case;
						END IF;
					END IF;
				end if;
			end if;
	end process;		

-------------------------------------------------------------------------
-- SDRAM Basic
-------------------------------------------------------------------------
	Addr <= wr_addr when S_cycle='1' else vAddr when V_cycle='1' else rAddr when R_cycle='1' else zAddr;
	dwrL <= NOT Addr(0);
	dwrU <= Addr(0);
	wr <= '0' when (R_cycle='1' AND rwr='0') OR (Z_cycle='1' AND zwr='0') OR S_cycle='1' else '1';
	refresh <= '0' when (vrefresh='0' AND V_cycle='1') else '1';
	datamux <= shift(8 downto 1)&shift(8 downto 1) when S_cycle='1' else datawr&datawr when R_cycle='1' else zdatawr&zdatawr;

	sd_ras <= isd_ras or sdcom_ena;
	sd_cas <= isd_cas or sdcom_ena;
	sd_we  <= isd_we  or sdcom_ena;
	sdaddr <= isdaddr;
	sdata <= datain when sdwrite='1' else "ZZZZZZZZZZZZZZZZ";

	process (sysclk, reset, datain, addr, casaddr, rd_int) begin
		if addr=casaddr and rd_int='0' then
			wrd_valid <='1';
		else	
			wrd_valid <='0';
		end if;	
		if reset = '0' then
			initstate <= (others => '0');
			init_done <= '0';
			sdram_state <= ph0;
			pass <= nop;
		else
			if (sysclk'event and sysclk='1') then
				sdcom_ena <= '1';
				ring <= "0000";
					case sdram_state is	--LATENCY=2
						when ph0 =>	sdram_state <= ph1;
									pass <= ras;
						
									ring(0) <= '1';
									IOB_clk <= '0';
									cpuclk <= '0';
									
						when ph1 =>	sdram_state <= ph2;
									sdcom_ena <= '0';
									ring(1) <= '1';
									IOB_clk <= '1';
									cpuclk <= '1';
									
						when ph2 =>	sdram_state <= ph3;
									pass <= cas;
									sdwrite <= '1';
									casaddr <= addr;
									datain <= datamux;
									IF refresh='0' then
										wr_int <= '1';
										rd_int <= '1';
									else
										wr_int <= wr ;
										rd_int <= '0';
									end if;
									ring(1) <= '1';
									IOB_clk <= '0';
									cpuclk <= '0';
									
						when ph3 =>	sdram_state <= ph4;
									sdcom_ena <= '0';
									ring(2) <= '1';
									IOB_clk <= '1';
									cpuclk <= '1';
									
						when ph4 =>	sdram_state <= ph5;
									pass <= nop;
									sdwrite <= '0';
									
									ring(2) <= '1';
									IOB_clk <= '0';
									cpuclk <= '0';
									
						when ph5 => sdram_state <= ph6;  
									ring(3) <= '1';
									IOB_clk <= '1';
									cpuclk <= '1';
									
						when ph6 =>	sdram_state <= ph7;
									ring(3) <= '1';
									IOB_clk <= '0';
									cpuclk <= '0';
									
						when ph7 =>	sdram_state <= ph0;
									if initstate /= "1111" then
										initstate <= initstate+1;
									else
										init_done <='1';	
									end if;
									ring(0) <= '1';
									IOB_clk <= '1';
									cpuclk <= '1';
									
						when others =>	null;
					end case;	
			end if;
		end if;	
	end process;		


	
-------------------------------------------------------------------------
-- SDRAM Init
-------------------------------------------------------------------------
	process (initstate, pass, addr,refresh, wr, datain, init_done, Latency, casaddr, rd_int, wr_int, dwrU, dwrL) begin
		isdaddr <= addr(21 downto 10);
		isd_ras <= '1';
		isd_cas <= '1';
		isd_we <= '1';
		ba <= Addr(23 downto 22);
			if wr_int='0' then
				dqm <= (not dwrU& not dwrL);
			else
				dqm <= "00";
			end if;	
			if init_done='0' then
				if pass=ras then
					case initstate is
						when "0010" => --PRECHARGE
							isdaddr(10) <= '1'; 	--all banks
							isd_ras <= '0';
							isd_cas <= '1';
							isd_we <= '0';
						when "0011"|"0100"|"0101"|"0110"|"0111"|"1000"|"1001"|"1010"|"1011"|"1100" => --AUTOREFRESH
							isd_ras <= '0';
							isd_cas <= '0';
							isd_we <= '1';
						when "1101" => --LOAD MODE REGISTER
							isd_ras <= '0';
							isd_cas <= '0';
							isd_we <= '0';
							ba <= "00";
							isdaddr <= "001000100010"; --BURST=4 LATENCY=2
	--						isdaddr <= "001000100000"; --BURST=1 LATENCY=2
						when others =>	null;	--NOP
					end case;
				end if;	
			else		
				case pass is
					when ras => --ACTIVE
						isd_ras <= '0';
						isd_we <= '1';
						if refresh = '0' then
							isd_cas <= '0'; --AUTOREFRESH
						else
							isd_cas <= '1'; --ACTIVE
						end if;	
					when cas => --READ or Write
						ba <= casaddr(23 downto 22);
						isdaddr <= '0' & '1' & '0' & casaddr(9 downto 1);--auto precharge
						isd_ras <= '1';
						isd_cas <= rd_int;
						isd_we  <= wr_int;
					when others =>		--NOP
				end case;
			end if;	
	end process;	

-----------------------------------------------------------------
-- store Data
-----------------------------------------------------------------
	reset_out <= init_done AND romready;
process(reset, sysclk)
begin
	if reset='0' then
		shift <= "011111111";
		mdcdd <= "00000000";
		mdbd <= "00000000";
		dma <= '0';
		dmad <= '0';
		romready <= '0';
	elsif sysclk'event and sysclk = '1' then
		mdbd <= mdb;
		mdcdd(5) <= not mdbd(5);
		dmad <= dma;
		IF dmad='1' AND dma='0' THEN
			romready <= '1';
		END IF;
		if mdcdd(5)='1' and mdbd(5)='1' then
			dma <= mdb(0);
			if shift(0)='0' OR dma='0' THEN
				shift <= mdbd(4)&"01111111";
			else	
				shift <= mdbd(4)&shift(8 downto 1);
			end if;	
		end if;
	end if;
end process; 
	
end;
