------------------------------------------------------------------------------
------------------------------------------------------------------------------
--                                                                          --
-- Copyright (c) 2009 Tobias Gubener                                        -- 
-- Subdesign CPC T-REX by TobiFlex                                          --
--                                                                          --
-- This source file is free software: you can redistribute it and/or modify --
-- it under the terms of the GNU General Public License as published        --
-- by the Free Software Foundation, either version 3 of the License, or     --
-- (at your option) any later version.                                      --
--                                                                          --
-- This source file is distributed in the hope that it will be useful,      --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of           --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            --
-- GNU General Public License for more details.                             --
--                                                                          --
-- You should have received a copy of the GNU General Public License        --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.    --
--                                                                          --
------------------------------------------------------------------------------
------------------------------------------------------------------------------

--000000-07FFFF  512K DK RAM Shadow
--080000-08FFFF  64K RAM	Shadow
--090000-093FFF  Kernalrom
--094000-097FFF  Kernalrom Shadow
--
--0A0000-0BFFFF  8x16K ROM Shadow
--0C0000-0FFFFF  16x16K ROM
--100000-1FFFFF  DSK A
--200000-2FFFFF  DSK B
--(200000-2FFFFF  DSK A)
--(300000-3FFFFF  DSK B)
--
--400000-47FFFF  512K DK RAM
--480000-48FFFF  64K RAM

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity shadowglue is
   port(        
		n_reset					: in std_logic;
		sysclk					: in std_logic;		--96MKHz
		cpuclk					: in std_logic;		--48MKHz
		clkena					: in std_logic;
		data					: in std_logic_vector(7 downto 0);
		addr					: in std_logic_vector(23 downto 0);
		n_iowr					: in std_logic;
		flat_mem				: in std_logic;
		intack					: in std_logic;
		mem_wr					: in std_logic;
		pal						: in std_logic;
		crtclk					: in std_logic;
		enable					: in std_logic;
		vdata					: in std_logic_vector(7 downto 0);
		Vsync					: buffer std_logic;
		vaddr					: out std_logic_vector(23 downto 0);
		sdaddr					: out std_logic_vector(23 downto 0);
		int						: out std_logic;
		pixclk					: out std_logic;
		CPCHsync				: out std_logic;
		CPCVsync				: out std_logic;
		CPCVideo				: out std_logic_vector(15 downto 0)
    );
end shadowglue;

ARCHITECTURE logic OF shadowglue IS

	COMPONENT crtc
    PORT (
			MA     : out STD_LOGIC_VECTOR (13 downto 0);
			RA     : out STD_LOGIC_VECTOR (4 downto 0);
			HSYNC  : out STD_LOGIC;
			VSYNC  : out STD_LOGIC;
			DISP   : buffer STD_LOGIC;
			CPCHSYNC  : out STD_LOGIC;
			CPCVSYNC  : out STD_LOGIC;
			CPCDISP   : out STD_LOGIC;
			CPCBLACK  : out STD_LOGIC;
			IOCLK  : in STD_LOGIC;
			RS     : in STD_LOGIC;
			CS     : in STD_LOGIC;
			RW     : in STD_LOGIC;
			Din    : in STD_LOGIC_VECTOR (7 downto 0);
			RESET  : in STD_LOGIC;
			CLK    : in STD_LOGIC;
			PAL    : in STD_LOGIC
        );
	END COMPONENT;
	
	
SIGNAL  intcnt     : std_logic_vector(5 downto 0);
SIGNAL  intcntp    : std_logic_vector(5 downto 0);
SIGNAL  vsyncd     : std_logic_vector(1 downto 0);
SIGNAL  interrupt  : std_logic;
SIGNAL  inter      : std_logic;
SIGNAL  intres     : std_logic;
SIGNAL  iflag      : std_logic;

SIGNAL  dktronik   : std_logic_vector(5 downto 0);
SIGNAL  dktroniksel: std_logic_vector(2 downto 0);
SIGNAL  dktab      : std_logic_vector(4 downto 0);
SIGNAL  romsel 	   : std_logic_vector(3 downto 0);
SIGNAL  MF	 	   : std_logic_vector(3 downto 0);

SIGNAL  coladdr	   : std_logic_vector(4 downto 0);


SIGNAL MA     : STD_LOGIC_VECTOR (13 downto 0);
SIGNAL RA     : STD_LOGIC_VECTOR (4 downto 0);
SIGNAL Hsync      : std_logic;
SIGNAL CPCDisp      : std_logic;
SIGNAL CPCDispd      : std_logic;
SIGNAL CPCBlack      : std_logic;
SIGNAL pixcnt     : STD_LOGIC_VECTOR (6 downto 0);
SIGNAL pixindex     : STD_LOGIC_VECTOR (4 downto 0);
SIGNAL colornr     : STD_LOGIC_VECTOR (4 downto 0);
SIGNAL n_video     : STD_LOGIC_VECTOR (5 downto 0);
SIGNAL MFH     : STD_LOGIC_VECTOR (1 downto 0);
SIGNAL CPCHsynci	: std_logic;
SIGNAL CPCVsynci	: std_logic;


type coltab_t is array(0 to 16) of std_logic_vector(4 downto 0);
signal coltab	  : coltab_t;

BEGIN  

--****************************
--IO Setupregister
--****************************
PROCESS (n_reset, cpuclk)
	BEGIN
		IF n_reset='0' THEN
			dktronik <= "000000";
			intres <= '0';
			romsel <= "0000";
			MF <= "0000";
		ELSIF rising_edge(cpuclk) THEN
        	IF clkena='1' THEN
			
--****************************
--MF2-Reg
--****************************
				intres <= '0';
				IF addr(15)='0' AND n_iowr='0' THEN 
					IF data(7 downto 6)="00" THEN 			--coloradresse
						IF data(4)='1' THEN
							coladdr <= "10000";
						ELSE
							coladdr <= data(4 downto 0);
						END IF;
					END IF;
					IF data(7 downto 6)="01" THEN 			--colorwert
						coltab(conv_integer(coladdr)) <= data(4 downto 0);
					END IF;
					IF data(7 downto 6)="10" THEN 	--AND intcnt/=0 AND THEN
						intres <= data(4);			--MF bit4=H => int reset
						MF <= data(3 downto 0);
					END IF;
			--DKtronik RAM
					IF data(7 downto 6)="11" THEN	--MF2-Reg
						dktronik <= data(5 downto 0);		--512KB
--						dktronik <= "00"&data(3 downto 0);	--128KB
--						dktronik <= "000"&data(2 downto 0);	--64KB
					END IF;
				END IF;

--****************************
--ROMSELECT
--****************************
				IF addr(13)='0' AND n_iowr='0' THEN			--IOWR ROM selektieren
					IF data(7 downto 3)="00000" THEN
						romsel <= data(3 downto 0);  		--16 roms
					ELSE		
						romsel <= "0000";  
					END IF;	
				END IF;				
			END IF;
		END IF;	
	END PROCESS;

--****************************
--Video controll 
--****************************
VC6845: crtc
	PORT MAP (
			MA => MA,    		--: out STD_LOGIC_VECTOR (13 downto 0);
			RA => RA,     		--: out STD_LOGIC_VECTOR (4 downto 0);
			HSYNC => Hsync,  	--: out STD_LOGIC;
			VSYNC => Vsync,  	--: out STD_LOGIC;
			DISP => CPCDisp,  		--: buffer STD_LOGIC;
			CPCHSYNC => CPCHsynci, --: out STD_LOGIC;
			CPCVSYNC => CPCVsynci, --: out STD_LOGIC;
			CPCDISP => open,  	--: out STD_LOGIC;
			CPCBLACK => CPCBlack,  	--: out STD_LOGIC;
			IOCLK => cpuclk,  	--: in STD_LOGIC;
			RS => addr(8),    	--: in STD_LOGIC;
			CS => (addr(14) or addr(9)),     --: in STD_LOGIC;
			RW =>  NOT clkena OR n_iowr,     		--: in STD_LOGIC;
			Din => data,   		--: in STD_LOGIC_VECTOR (7 downto 0);
			RESET => n_reset, 	--: in STD_LOGIC;
			CLK => crtclk, 		--: in STD_LOGIC;
			PAL => pal   		--: in STD_LOGIC
        );
        
        
PROCESS (n_reset, sysclk, cpuclk, enable, MA, RA, CPCHsynci, CPCVsynci, pixcnt, vdata, colornr, Hsync, CPCDispd, MFH)
	BEGIN
		IF enable='1' THEN
			vaddr(23 downto 16) <= "00001000";	--08 shadow crt memory
			vaddr(15 downto 14) <= MA(13 downto 12);
			vaddr(13 downto 11) <= RA(2 downto 0);
			vaddr(10 downto 1)  <= MA(9 downto 0);
			vaddr(0)  <= NOT pixcnt(6);
			CPCHsync <= CPCHsynci;
			CPCVsync <= CPCVsynci;
		ELSE
			vaddr <= (OTHERS=>'0');
			CPCHsync <= '0';
			CPCVsync <= '0';
		END IF;
	
		pixclk <= pixcnt(1);
		CASE colornr IS
 		        WHEN "00000" => n_video <= "010101";
 		        WHEN "00001" => n_video <= "011101";
 		        WHEN "00010" => n_video <= "001101";
 		        WHEN "00011" => n_video <= "111101";

 		        WHEN "00100" => n_video <= "000001";
 		        WHEN "00101" => n_video <= "110001";
 		        WHEN "00110" => n_video <= "000101";
 		        WHEN "00111" => n_video <= "110101";

 		        WHEN "01000" => n_video <= "110100";
 		        WHEN "01001" => n_video <= "111111";
 		        WHEN "01010" => n_video <= "111100";
 		        WHEN "01011" => n_video <= "111111";

 		        WHEN "01100" => n_video <= "110000";
 		        WHEN "01101" => n_video <= "110011";
 		        WHEN "01110" => n_video <= "110100";
 		        WHEN "01111" => n_video <= "110111";

 		        WHEN "10000" => n_video <= "000001";
 		        WHEN "10001" => n_video <= "001101";
 		        WHEN "10010" => n_video <= "001100";
 		        WHEN "10011" => n_video <= "001111";

		        WHEN "10100" => n_video <= "000000";
 		        WHEN "10101" => n_video <= "000011";
 		        WHEN "10110" => n_video <= "000100";
 		        WHEN "10111" => n_video <= "000111";

 		        WHEN "11000" => n_video <= "010001";
 		        WHEN "11001" => n_video <= "011101";
 		        WHEN "11010" => n_video <= "011100";
 		        WHEN "11011" => n_video <= "011111";

		        WHEN "11100" => n_video <= "010000";
 		        WHEN "11101" => n_video <= "010011";
 		        WHEN "11110" => n_video <= "010100";
 		        WHEN OTHERS  => n_video <= "010111";
 		END CASE; 

		IF falling_edge(Hsync) THEN
			MFH <= MF(1 downto 0);
		END IF;
		
		IF CPCDispd='0' THEN
			pixindex <= "10000";				--border
		ELSE	
			pixindex <= "00000"; 		
			CASE MFH IS 
				WHEN "01" =>
					CASE pixcnt(5 downto 4) IS
						WHEN "00" => pixindex(0) <= vdata(7);
									 pixindex(1) <= vdata(3);  
						WHEN "01" => pixindex(0) <= vdata(6);  
									 pixindex(1) <= vdata(2);  
						WHEN "10" => pixindex(0) <= vdata(5);  
									 pixindex(1) <= vdata(1);  
						WHEN OTHERS => pixindex(0) <= vdata(4);  
									 pixindex(1) <= vdata(0);  
					END CASE;
				WHEN "10" =>
					CASE pixcnt(5 downto 3) IS
						WHEN "000" => pixindex(0) <= vdata(7);
						WHEN "001" => pixindex(0) <= vdata(6);  
						WHEN "010" => pixindex(0) <= vdata(5);  
						WHEN "011" => pixindex(0) <= vdata(4);  
						WHEN "100" => pixindex(0) <= vdata(3);  
						WHEN "101" => pixindex(0) <= vdata(2);  
						WHEN "110" => pixindex(0) <= vdata(1);  
						WHEN OTHERS => pixindex(0) <= vdata(0);  
					END CASE;
				WHEN OTHERS =>
					CASE pixcnt(5) IS
						WHEN '0' =>  pixindex(0) <= vdata(7);
									 pixindex(2) <= vdata(5);  
									 pixindex(1) <= vdata(3);  
									 pixindex(3) <= vdata(1);  
						WHEN OTHERS => pixindex(0) <= vdata(6);  
									 pixindex(2) <= vdata(4);  
									 pixindex(1) <= vdata(2);  
									 pixindex(3) <= vdata(0);  
					END CASE;
			END CASE;
		END IF;
 		
 		              	
		IF n_reset='0' THEN
			pixcnt <= "1000000";
		ELSIF rising_edge(sysclk) THEN
			IF pixcnt(1 downto 0)="10" THEN
				pixcnt <= pixcnt+2;
				IF pixcnt(5 downto 2)="1111" THEN
					CPCDispd <= CPCDisp;
				END IF;	
			ELSE
				pixcnt <= pixcnt+1;
			END IF;
		
			colornr <= coltab(conv_integer(pixindex(4 downto 0)));
			if hsync='1' OR vsync='1' OR CPCBlack='1' OR enable='0' then
				CPCVideo <= "0000000000000000";
			else
				CPCVideo <= (n_video(1)& n_video(0)& n_video(0)& n_video(0)& n_video(0)& n_video(3)& n_video(2)& n_video(2)& n_video(2)& n_video(2)& '0' & n_video(5)& n_video(4)& n_video(4)& n_video(4)& n_video(4));
			end if;
		END IF;	
	END PROCESS;

--****************************
--RAM/ROM Banking
--****************************
PROCESS (cpuclk, dktroniksel, dktronik, mem_wr, addr, romsel, dktab, flat_mem, MF)
	BEGIN
--DKtronik RAM
		dktab <= dktronik(2 downto 0)&addr(15 downto 14);
		CASE dktab IS
 		        WHEN "00000"        =>      dktroniksel <= "000";
 		        WHEN "00001"        =>      dktroniksel <= "001";
 		        WHEN "00010"        =>      dktroniksel <= "010";
 		        WHEN "00011"        =>      dktroniksel <= "011";

 		        WHEN "00100"        =>      dktroniksel <= "000";
 		        WHEN "00101"        =>      dktroniksel <= "001";
 		        WHEN "00110"        =>      dktroniksel <= "010";
 		        WHEN "00111"        =>      dktroniksel <= "111";

 		        WHEN "01000"        =>      dktroniksel <= "100";
 		        WHEN "01001"        =>      dktroniksel <= "101";
 		        WHEN "01010"        =>      dktroniksel <= "110";
 		        WHEN "01011"        =>      dktroniksel <= "111";

 		        WHEN "01100"        =>      dktroniksel <= "000";
 		        WHEN "01101"        =>      dktroniksel <= "011";
 		        WHEN "01110"        =>      dktroniksel <= "010";
 		        WHEN "01111"        =>      dktroniksel <= "111";

 		        WHEN "10000"        =>      dktroniksel <= "000";
 		        WHEN "10001"        =>      dktroniksel <= "100";
 		        WHEN "10010"        =>      dktroniksel <= "010";
 		        WHEN "10011"        =>      dktroniksel <= "011";

		        WHEN "10100"        =>      dktroniksel <= "000";
 		        WHEN "10101"        =>      dktroniksel <= "101";
 		        WHEN "10110"        =>      dktroniksel <= "010";
 		        WHEN "10111"        =>      dktroniksel <= "011";

 		        WHEN "11000"        =>      dktroniksel <= "000";
 		        WHEN "11001"        =>      dktroniksel <= "110";
 		        WHEN "11010"        =>      dktroniksel <= "010";
 		        WHEN "11011"        =>      dktroniksel <= "011";

		        WHEN "11100"        =>      dktroniksel <= "000";
 		        WHEN "11101"        =>      dktroniksel <= "111";
 		        WHEN "11110"        =>      dktroniksel <= "010";
 		        WHEN "11111"        =>      dktroniksel <= "011";
 		        
 		        WHEN OTHERS         =>   null;
		END CASE;

		sdaddr(23 downto 0) <= addr(23 downto 0);	
		IF flat_mem='0' THEN
	--DKtronik + 6128 RAM
			sdaddr(15 downto 14) <= dktroniksel(1 downto 0);
			IF dktroniksel(2)='1' THEN
				sdaddr(23 downto 16) <= ("00000"& dktronik(5 downto 3)); --DK ram 000000-07FFFF 
			ELSE
				sdaddr(23 downto 16) <=  "00001000";		--64K RAM 080000-08FFFF
			END IF;
	--ROMs		
			IF mem_wr='1' THEN
				IF addr(15 downto 14)="11" AND MF(3)='0' THEN
					sdaddr(23 downto 14) <= ("0000101"&romsel(2 downto 0));	--ROM 0-7 0A0000-0BFFFF 
				END IF;
				IF addr(15 downto 14)="00" AND MF(2)='0' THEN
					sdaddr(23 downto 14) <= B"0000100101"; 	--Kernal 094000-097FFF
				END IF;
			END IF;
		END IF;
	END PROCESS;

--****************************
--CPC Interrupt generation
--****************************

int <= NOT inter;
PROCESS (interrupt, intres, intack)
	BEGIN
		IF intres='1' OR intack='1' THEN
			inter <= '0';
		ELSIF rising_edge(interrupt) THEN
			inter <= '1';
		END IF;
	END PROCESS;

PROCESS (Hsync, intres, intack, intcnt)
	BEGIN
		intcntp <= intcnt+1;
		IF intres='1' OR intack='1' THEN
			interrupt <= '0';
			IF intres='1' THEN
				intcnt <= "000000";
			END IF;
			IF intack='1' THEN
				intcnt(5) <= '0';
			END IF;
		ELSIF falling_edge(Hsync) THEN
			vsyncd <=(vsyncd(0)& Vsync);
			IF intcnt=51 OR (vsyncd(1)='0' AND  vsyncd(0)='1')THEN
				interrupt <= intcntp(5);-- OR interrupt;
				intcnt <= "000000";
			ELSE
				interrupt <= '0';
				intcnt <= intcntp;
			END IF;
		END IF;
	END PROCESS;
END;	